#define USE_OLD_TONEMAP

#include "shaders/common/gbuffer.h"

#include "shaders/common/hlsl.h"
#include "shaders/common/bng.hlsl"
// #include "../postFx.hlsl"
#include "shaders/common/postfx/hdr/tonemap.h.hlsl"

Texture2D<float4> backbuffer : register(t0);

// luminanceTex
// size [1][1]
Texture2D<float4> adaptedLum : register(t1);
// bloomTex
Texture2D<float4> bloomFinal : register(t2);

// The 255x1 color correction ramp texture for doing full screen color correction.
// colorCorrectionTex
Texture1D<float4> colorCorrectionRamp : register(t3);

RWTexture2D<float4> destTex : register(u0);

// In it's own buffer since it might change during runtime while all the other settings will be set once at start.
cbuffer RuntimeSettingsCBuffer : register(b0)
{
    // The blue shift color value.
    // float3 factorHSL = float3(1, 1, 1);
    // float enabledHSL = 1.0f;
    // float4 HSL = float4(factorHSL.x, factorHSL.y, factorHSL.z, enabledHSL);
    float4 HSL;                     // float4(1, 1, 1, 1);

    float enabled;                  // 1.0f

    // Blends between the scene and the blue shifted version
    // of the scene for a cinematic desaturated night effect.
    float enableBlueShift;          // 0.0f

    // Blends between the scene and the bloomed scene.
    float bloomScale;               // 1.0f

    float oneOverGamma;             // 1.0f / gamma
    float colorCorrectionStrength;  // 1.0f

    float maxAdaptedLum;            // 1.0f

    // The tone mapping middle grey or exposure value used
    // to adjust the overall "balance" of the image.
    float middleGray;               // 0.5f

    float blueShiftLumVal;          // 0.1f
    float3 blueShiftColor;          // Point3F(1.05, 0.97, 1.27);

    uint colorCorrectionRampDefaultWidth;
    uint colorCorrectionRampWidth;
};

uint2 getRtDimensions() {
    uint2 dim;
    destTex.GetDimensions(dim.x, dim.y);
    return dim;
}

SamplerState samplerLinear
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Wrap;
    AddressV = Wrap;
    AddressW = Wrap;
};

// Define the number of threads to be executed in a single thread group when we dispatch the compute shader.
[numthreads(8, 8, 1)]
void main(uint3 dispatchThreadID : SV_DispatchThreadID)
{
    const uint2 storePos = uint2(dispatchThreadID.xy);
    static float2 rtSizeInv = 1.0f / getRtDimensions();
    float2 uv = (storePos + 0.5) * rtSizeInv;

    float adaptedLumVal;

    float4 sampleSRGB;
    {
        float4 sample = backbuffer[storePos];
        if (enabled == 0.0f) {
            destTex[storePos] = sample;
            return;
        }

        adaptedLumVal = adaptedLum[uint2(0,0)].r;
        float4 bloom = bloomFinal.SampleLevel(samplerLinear, uv, 0);

        // For very low light conditions, the rods will dominate the perception
        // of light, and therefore color will be desaturated and shifted
        // towards blue.
        if (enableBlueShift > 0.0f) {
            const float3 LUMINANCE_VECTOR = float3(0.2125f, 0.7154f, 0.0721f);

            // Define a linear blending from -1.5 to 2.6 (log scale) which
            // determines the lerp amount for blue shift

            float coef = 1.0f - (adaptedLumVal / blueShiftLumVal);
            coef = saturate(coef * enableBlueShift);

            // Lerp between current color and blue, desaturated copy
            float3 rodColor = dot(sample.rgb, LUMINANCE_VECTOR) * blueShiftColor;
            sample.rgb = lerp(sample.rgb, rodColor, coef);

            rodColor = dot(bloom.rgb, LUMINANCE_VECTOR) * blueShiftColor;
            bloom.rgb = lerp(bloom.rgb, rodColor, coef);
        }
		
        if (maxAdaptedLum > 1.0f) {
            sample *= min(maxAdaptedLum, middleGray / adaptedLumVal);
        }

        // Add the bloom effect.
        sample += bloomScale * bloom * min(maxAdaptedLum, middleGray / adaptedLumVal);

        // static bool in bge.hlsl
        if(enableHDR) {
            sample.rgb = tonemap(sample.rgb);

            // HSL color correction.
            if(HSL.a) {
                sample.rgb = rgb2hsv(sample.rgb) * HSL.rgb;
                sample.rgb = hsv2rgb(sample.rgb);
            }

            // To gamma space.
            sample.rgb = linearToGammaColor(sample.rgb);
        }

        sampleSRGB = sample;
    }

    // Clamp color correction strength value.
    float colorCorrectionStrengthValue = saturate(colorCorrectionStrength);

    // Get ramp pixel position based on channel value and ramp width.
    uint rampR = sampleSRGB.r * (colorCorrectionRampWidth - 1);
    uint rampG = sampleSRGB.g * (colorCorrectionRampWidth - 1);
    uint rampB = sampleSRGB.b * (colorCorrectionRampWidth - 1);

    // Apply the color correction.
    if(colorCorrectionStrengthValue > 0) {
        sampleSRGB.r = lerp(sampleSRGB.r, colorCorrectionRamp[rampR].r, colorCorrectionStrengthValue);
        sampleSRGB.g = lerp(sampleSRGB.g, colorCorrectionRamp[rampG].g, colorCorrectionStrengthValue);
        sampleSRGB.b = lerp(sampleSRGB.b, colorCorrectionRamp[rampB].b, colorCorrectionStrengthValue);
    }

    // Apply gamma correction
    sampleSRGB.rgb = pow(abs(sampleSRGB.rgb), oneOverGamma);

    // Copy color from sample to output texture.
    destTex[storePos] = sampleSRGB;
}